#!/usr/bin/env python3

import itertools

import numpy as np
import scipy.spatial

import lapjv

from .common import ConfigurationError
from .dtgw_tools import NullObserver, info, shortest_warp_path
from .config import CONFIG


def compute_dtgw(features1, features2, eps, metric=scipy.spatial.distance.cdist, observer=NullObserver()):
	"""Computes the aprroximated dtgw distance between two temporal graphs.
	
	Arguments:
	----------
	features1 : numpy.ndarray of shape (l1, n1, k)
		Features of the first graph.
		features1[i,j,:] is the feature vector
		of vertex j in layer i of the first temporal graph
	features2 : numpy.ndarray of shape (l2, n2, k)
		Analogous to features1
	eps : numpy.ndarray of shape (k)
		Feature vector of a nonexisting (deleted) vertex.
	metric :
		Metric function to compute distances
	observer : 
		Observer object to notify on interal state changes

	Returns:
	---------
	d : float
		The approximate dtgw distance as determined by the alternating heuristic.
	"""
	t1, n1, k1 = features1.shape
	t2, n2, k2 = features2.shape
	if k1 != k2:
		raise ValueError("Arguments have wrong dimensions")
	k = k1
	n = max(n1, n2)

	features1 = np.hstack((features1, np.broadcast_to(eps, (t1, n-n1, k))))
	features2 = np.hstack((features2, np.broadcast_to(eps, (t2, n-n2, k))))

	warp_path = None
	if CONFIG.get("heuristic initialization") == "diagonal warping":
		warp_path = list(shortest_warp_path(t1, t2))
	elif CONFIG.get("heuristic initialization") == "optimistic warping":
		warp_path = list(optimistic_warp_path(t1, t2, features1, features2, metric))
	elif CONFIG.get("heuristic initialization") == "sigma*":
		warp_path = list(itertools.product(range(t1), range(t2)))
	elif CONFIG.get("heuristic initialization") == "optimistic matching":
		matchcost = np.zeros((n, n))
		for i in range(t1):
			tmp = np.stack(
				(
					metric(features1[i,:,:], features2[j,:,:])
					for j in range(t2)
				),
				axis = -1
			)
			matchcost += tmp.min(axis=-1)
	else:
		raise ConfigurationError('Invalid configuration for "heuristic initialization"')
	prev_cost = float('inf')
	iteration = 0
	while True:
		iteration += 1
		if warp_path is not None:
			matchcost = sum(
				metric(features1[i,:,:], features2[j,:,:])
				for i,j in warp_path
			)
		matching, _, _ = lapjv.lapjv(matchcost)

		warpcost = sum(
			metric(features1[:,v,:], features2[:,w,:])
			for v,w in enumerate(matching)
		)
		cost, warp_path = dynamic_timewarping(warpcost)
		observer.update(
			iteration = iteration,
			cost = cost,
			vertex_matching = (
				(
					v if v < n1 else None,
					w if w < n2 else None
				)
				for (v, w) in enumerate(matching)
			),
			warp_path = warp_path,
			final = cost >= prev_cost
		)
		if cost >= prev_cost:
			break
		prev_cost = cost

		info("Iteration: {}, Cost: {}".format(iteration, cost))
	
	return cost


def optimistic_warp_path(t1, t2, features1, features2, metric):
	cost_matrix = np.empty((t1, t2))
	for i, j in np.ndindex(t1, t2):
		cost_matrix[i,j] = lapjv.lapjv(metric(features1[i,:], features2[j,:]))[2][0]
	return dynamic_timewarping(cost_matrix)[1]


def dynamic_timewarping(cost_matrix):
	"""Computes the dynamic timewarping distance between two sequences A and B
	where the distance between A[i] and B[j] is given by »cost_matrix[i,j]«.

	Returns:
	--------
	d : number
		The dynamic timewarping distance
	path : [np.array((x, y))]) 
		An optimal warping path as list of coordinate pairs
	"""
	n1, n2 = cost_matrix.shape
	matrix = np.full((n1+1, n2+1), np.inf)
	# the last row and column serve as guard rails
	matrix[:n1,:n2] = cost_matrix
	# Compute the warp distance of seq1[i:] and seq2[j:] and store
	# that value in matrix[i,j]:
	matrix[n1,n2] = 0
	for i in range(n1-1, -1, -1):
		for j in range(n2-1, -1, -1):
			matrix[i,j] += min(
				matrix[i+1, j+1],
				matrix[i,   j+1],
				matrix[i+1, j  ]
			)
	# build warp path from (0, 0) to (n1-1, n2-1):
	path = [np.array((0, 0))]
	indices = np.array([
		(1, 1),
		(0, 1),
		(1, 0)
	])
	while np.any(path[-1] != (n1-1, n2-1)):
		path.append(path[-1] + indices[np.argmin([
			matrix[tuple(path[-1] + index)] for index in indices
		])])
	return matrix[0,0], path
