import itertools
import collections

from .config import CONFIG


class NullObserver:
	def update(self, **kwargs):
		pass


class LoggingObserver:
	def __init__(self, outfile, vertex_map_1=str, vertex_map_2=str, layer_map_1=str, layer_map_2=str, final_only=True):
		"""Logfile-writing dtgw observer

		Arguments:
		----------
		outfile :
			File to write to
		vertex_map_1, vertex_map_2:
			Functions to apply to any vertex from the first resp. second
			temporal graph before outputting
		layer_map_1, layer_map_2:
			Functions to apply to any layer of the first resp. second
			temporal graph before outputting
		final_only:
			Whether to log only the final result (True) or intermediate
			states, too (False)
		"""
		self._outfile = outfile
		self._final_only = final_only
		self._vertex_map_1 = vertex_map_1
		self._vertex_map_2 = vertex_map_2
		self._layer_map_1 = layer_map_1
		self._layer_map_2 = layer_map_2

	def update(self, vertex_matching, warp_path, final, cost, iteration=None, **kwargs):
		if not final and self._final_only:
			return
		if not self._final_only:
			self._outfile.write("============================\n\n")
		if iteration is not None:
			self._outfile.write("Iteration = {}\n".format(iteration))
		self._outfile.write("Cost = {}\n\n".format(cost))
		self._outfile.write(print_vertex_matching(vertex_matching, self._vertex_map_1, self._vertex_map_2))
		self._outfile.write("\n")
		self._outfile.write(print_warping(warp_path, self._layer_map_1, self._layer_map_2, abbrev=True))


def info(msg):
	if CONFIG["verbosity"] >= 1:
		print(msg)


def print_vertex_matching(matching, vertex_map_1=str, vertex_map_2=str):
	"""Print vertex matching as a human-readable string."""
	output = ""
	for v, w in matching:
		output += str(vertex_map_1(v)) + " <=> " + str(vertex_map_2(w)) + "\n"
	return output


def last(iterable):
	return collections.deque(iterable, maxlen=1).pop()


def print_warping(warp_path, layer_map_1=str, layer_map_2=str, abbrev=False):
	"""Print warping path as a human readable string.
	If »abbrev« is True, an abbreviated format is used that omits trivial parts.
	"""
	output = ""
	def p(x):
		nonlocal output
		output += str(layer_map_1(x[0])) + " --- " + str(layer_map_2(x[1])) + "\n"
	if abbrev:
		for delta, it in itertools.groupby(
			warp_path,
			lambda x: x[0]-x[1]
		):
			p(next(it))
			try:
				tmp = next(it)
				try:
					tmp = last(it)
					output += "[...]\n"
				except IndexError:
					pass  # »it« only had two items
				p(tmp)
			except StopIteration:
				pass  # »it« only had one item
	else:
		for x in warp_path:
			p(x)
	return output


def shortest_warp_path(n, m):
	i, j = 0, 0
	while i < n and j < m:
		yield i, j
		a = (i+1) * m
		b = (j+1) * n
		if a < b + n:
			i += 1
		if b < a + m:
			j += 1
