import argparse
import time

from .common import read_timeedges
from .dtgw_tools import NullObserver, LoggingObserver, info
from .config import CONFIG
from .vertex_features import FEATURES


def main():
	parser = argparse.ArgumentParser()
	parser.add_argument("graph1", type=argparse.FileType("r"))
	parser.add_argument("graph2", type=argparse.FileType("r"))
	parser.add_argument("--verbose", "-v", action="count", default=0)
	parser.add_argument("--log", type=argparse.FileType("w"), help="Logfile")
	parser.add_argument("--ilp", action="store_true", help="Use ILP instead of heuristic")
	parser.add_argument("--features", choices=FEATURES.keys(), default="degrees")
	args = parser.parse_args()

	if args.verbose:
		CONFIG["verbosity"] = int(args.verbose)

	if args.ilp:
		from . import dtgw_ilp
		compute_dtgw = dtgw_ilp.compute_dtgw
	else:
		from . import dtgw_alternating
		compute_dtgw = dtgw_alternating.compute_dtgw

	dtgw(args.graph1, args.graph2, compute_dtgw, FEATURES[args.features], args.log)


def dtgw(edgefile1, edgefile2, compute_dtgw, feature, logfile=None):
	"""Compute the dtgw distance between two temporal graphs
	and prints the result to stdout.

	Arguments
	---------
	edgefile1, edgefile2:
		Time-edge files for the two temporal graphs
	compute_dtgw:
		one of [dtgw_ilp.compute_dtgw, dtgw_alternating.compute_dtgw]
		or any function with the same interface
	feature:
		A feature as described in the vertex_features module
	logfile: (optional)
		Logfile to write to
	"""
	start = time.time()
	tg1, layer_index1, vertex_index1 = read_timeedges(edgefile1)
	tg2, layer_index2, vertex_index2 = read_timeedges(edgefile2)

	if logfile:
		observer = LoggingObserver(
			outfile = logfile,
			vertex_map_1 = vertex_index1.lookup,
			vertex_map_2 = vertex_index2.lookup,
			layer_map_1 = layer_index1.lookup,
			layer_map_2 = layer_index2.lookup
		)
	else:
		observer = NullObserver()

	info("Graphs read. Elapsed time: {}s".format(time.time() - start))

	features1 = feature["compute"](tg1, layer_index1, vertex_index1)
	features2 = feature["compute"](tg2, layer_index2, vertex_index2)
	info("Features computed. Elapsed time: {}s".format(time.time() - start))
	distance = compute_dtgw(
		features1,
		features2,
		eps = feature["eps"],
		metric = feature["metric"],
		observer = observer
	)
	print("Result:", distance)
	info("Elapsed time: {}s".format(time.time() - start))


if __name__ == "__main__":
	main()