"""This module contains a selection of vertex features (aka signatures).
They are stored in the FEATURES dict under arbitrary string keys.

Each feature is a dict with three keys:
"compute" contains a function that is passed
	* a temporal graph with t layers and n vertices, 
	* a layer indexing, and
	* a vertex indexing
	and must return a t*n*k numpy array for some k>=1 containing
	the features for all vertices and layers
"eps" contains the feature vector for a nonexisting vertex
"metric" contains a function that computes the pairwise distances between two
	arrays of feature vectors (e.g. scipy.spatial.distance.cdist)
"""


import functools

import numpy as np
import scipy.spatial
import graph_tool as gt
import graph_tool.centrality


FEATURES = {}

FEATURES["degrees"] = {
	"compute": lambda tg, _1, _2: 
		degree_matrix(tg)[:,:,np.newaxis],
	"eps": 0,
	"metric": scipy.spatial.distance.cdist,
}

FEATURES["betweenness"] = {
	"compute": lambda tg, _1, _2: 
		betweenness_matrix(tg)[:,:,np.newaxis],
	"eps": 0,
	"metric": scipy.spatial.distance.cdist,
}

FEATURES["degrees,betweenness"] = {
	"compute": lambda tg, _1, _2: 
		np.stack([
				degree_matrix(tg),
				betweenness_matrix(tg)*100
			],
			axis=-1
		),
	"eps": [0, 0],
	"metric": functools.partial(scipy.spatial.distance.cdist, metric="cityblock"),
}

FEATURES["degrees,union-betweenness"] = {
	"compute": lambda tg, _1, _2: 
		np.stack(
			np.broadcast_arrays(
				degree_matrix(tg),
				10000 * gt.centrality.betweenness(tg.union_graph())[0].get_array()[np.newaxis,:]
			),
			axis=-1
		),
	"eps": [0, 0],
	"metric": functools.partial(scipy.spatial.distance.cdist, metric="cityblock"),
}


def degree_matrix(tgraph):
	"""Given a temporal graph, returns a matrix where entry
	(i, v) contains the degree of vertex v in layer i
	"""
	degrees = np.empty((tgraph.lifetime, tgraph.num_vertices()))
	vertices = tgraph.get_vertices()
	for i in range(tgraph.lifetime):
		degrees[i, :] = tgraph.layer(i).get_out_degrees(vertices)
	return degrees


def betweenness_matrix(tgraph):
	"""Given a temporal graph, returns a matrix where entry
	(i, v) contains the betweenness of vertex v in layer i
	"""
	betweenness = np.empty((tgraph.lifetime, tgraph.num_vertices()))
	vertices = tgraph.get_vertices()
	for i in range(tgraph.lifetime):
		vp, _ = gt.centrality.betweenness(tgraph.layer(i))
		betweenness[i,:] = vp.get_array()
	return betweenness
